import { Asset, assetManager, AssetManager, AudioClip, Prefab, resources, SceneAsset, Node, instantiate } from "cc";

/**
 * Resource Manager
 */
class ResourceManager {
    private static singleInstance: ResourceManager | null = null;
    static getInstance(): ResourceManager {
        if (ResourceManager.singleInstance == null) {
            ResourceManager.singleInstance = new ResourceManager();
        }
        return ResourceManager.singleInstance;
    }
    private _subfix: string = 'Ctrl';

    constructor() {

    }

    /**
     * 加载bundle
     * @param res bundle路径
     * @param version 版本，可选
     */
    public async loadBundle(res: string, version?: string,): Promise<AssetManager.Bundle> {
        return new Promise((resolve, reject) => {
            console.log('ResourceManager loadBundle() res is: ', res, ' version: ', version);
            let index = res.lastIndexOf('/');
            let name = res.substring(index + 1);
            let bundle = assetManager.getBundle(name);
            if (bundle) {
                if (version && version?.length > 0) {
                    assetManager.loadBundle(res, { version: version }, (err, bundle) => {
                        if (err) {
                            console.error(`loadBundle ${res}`);
                            return reject(err);
                        }
                        return resolve(bundle);
                    });
                }
                else {
                    assetManager.loadBundle(res, {}, (err, bundle) => {
                        if (err) {
                            console.error(`loadBundle ${res}`);
                            return reject(err);
                        }
                        return resolve(bundle);
                    });
                }
            }
            else {
                // @ts-ignore
                return resolve(bundle);
            }
        });
    }

    public async bundleRes(bundle: AssetManager.Bundle, res: string, type: typeof Asset) {
        return new Promise((resolve, reject) => {
            bundle.load(res, type, (err, result) => {
                if (err) {
                    console.error('load bundle error');
                    return reject(err);
                }
                return resolve(result);
            });
        });
    }

    public async bundleLoadPrefab(url: string, res: string, version?: string): Promise<Prefab> {
        let index = url.lastIndexOf('/');
        let name = url.substring(index + 1);
        let bundle: AssetManager.Bundle | null = assetManager.getBundle(name);
        if (bundle) {
            bundle = await this.loadBundle(url, version);
        }
        return new Promise((resolve, reject) => {
            bundle?.load(res, Prefab, (err, prefab) => {
                if (err) {
                    console.error('load bundle prefab error');
                    return reject(err);
                }
                return resolve(prefab);
            });
        });
    }

    public async bundleLoadAudioClip(url: string, res: string, version?: string): Promise<AudioClip> {
        let index = url.lastIndexOf('/');
        let name = url.substring(index + 1);
        let bundle: AssetManager.Bundle | null = assetManager.getBundle(name);
        if (bundle) {
            bundle = await this.loadBundle(url, version);
        }
        return new Promise((resolve, reject) => {
            bundle?.load(res, AudioClip, (err, result) => {
                if (err) {
                    console.error('load bundle audioclip error');
                    return reject(err);
                }
                return resolve(result);
            });
        });
    }

    public async bundleLoadScene(url: string, sceneName: string, version?: string): Promise<SceneAsset> {
        let index = url.lastIndexOf('/');
        let name = url.substring(index + 1);
        let bundle: AssetManager.Bundle | null = assetManager.getBundle(name);
        if (bundle) {
            bundle = await this.loadBundle(url, version);
        }
        return new Promise((resolve, reject) => {
            bundle?.loadScene(sceneName, (err, result) => {
                if (err) {
                    console.error('load bundle scene error');
                    return reject(err);
                }
                return resolve(result);
            });
        });
    }


    public async bundleLoadRes(url: string, res: string, type: typeof Asset, version?: string): Promise<Asset> {
        let index = url.lastIndexOf('/');
        let name = url.substring(index + 1);
        let bundle: AssetManager.Bundle | null = assetManager.getBundle(name);
        if (bundle) {
            bundle = await this.loadBundle(url, version);
        }
        return new Promise((resolve, reject) => {
            bundle?.load(res, type, (err, result) => {
                if (err) {
                    console.error('load bundle error');
                    return reject(err);
                }
                return resolve(result);
            });
        });
    }

    public async bundleLoadPrefabNode(url: string, res: string, version?: string): Promise<Node> {
        let index = url.lastIndexOf('/');
        let name = url.substring(index + 1);
        let bundle = assetManager.getBundle(name);
        if (bundle) {
            bundle = await this.loadBundle(url, version);
        }
        //
        let prefab = await this.bundleLoadPrefab(url, res, version);
        return new Promise((resolve, reject) => {
            let resIndex = res.lastIndexOf('/');
            let resName = res.substring(resIndex + 1);
            let node: Node | null = null;
            node = instantiate(prefab);
            node?.addComponent(resName + this._subfix);
            return resolve(node);
        });
    }

    /**
     * 以下接口没有暴露给绘本
     */
    /**
     * 加载场景
     * @param sceneUrl 
     */
    public async loadScene(sceneUrl: string): Promise<SceneAsset> {
        return new Promise((resolve, reject) => {
            resources.loadScene(sceneUrl, (err, scene) => {
                if (err) {
                    console.error(`loadPrefab ${sceneUrl} ${err}`);
                    return reject(err);
                }
                return resolve(scene);
            });
        });
    }

    /**
     * 加载预制体生成预制对象
     * @param res   resources目录下资源路径
     */
    public async loadPrefab(res: string): Promise<Prefab> {
        console.log('load prefab: ', res);
        return new Promise((resolve, reject) => {
            resources.load(res, Prefab, (err, prefab) => {
                if (err) {
                    console.error(`loadPrefab ${res} ${err}`);
                    return reject(err);
                }
                return resolve(prefab);
            });
        });
    }

    /**
     * 加载音频生成音频Clip
     * @param res   resources目录下资源路径
     */
    public async loadAudioClip(res: string): Promise<AudioClip> {
        return new Promise(async (resolve, reject) => {
            resources.load(res, AudioClip, (err, audioClip) => {
                if (err) {
                    console.error(`loadAudioClip ${res} ${err}`);
                    return reject(err);
                }
                resolve(audioClip);
            });
        });
    }

    public async loadRes(res: string, type?: Asset) {
        return new Promise((resolve, reject) => {
            // @ts-ignore
            resources.load(res, type, (err, asset) => {
                if (err) {
                    console.error(`loadRes ${res} ${err}`);
                    return reject(err);
                }
                resolve(asset);
            });
        });
    }

    /**
     * 加载预制体实例化预制对象节点
     * @param res   resources目录下资源路径
     */
    public async loadPrefabNode(res: string): Promise<Node> {
        console.log('load prefab node: ', res);
        let index = res.lastIndexOf('/');
        let name = res.substring(index + 1);
        return new Promise((resolve, reject) => {
            resources.load(res, Prefab, (err, prefab) => {
                let node: Node | null = null;
                if (err) {
                    console.error(`loadPrefabNode ${res} ${err}`);
                    return reject(err);
                } else {
                    node = instantiate(prefab);
                    node?.addComponent(name + this._subfix);
                }
                return resolve(node);
            });
        });
    }


}

export {
    ResourceManager
}
